@@ -17,16 +17,17 @@ use std::ops::{Index, IndexMut};
 use std::ptr::NonNull;
 use std::slice;
 
+#[warn(missing_docs)]
 /**
 A queue that also allows direct index addressing.
 
 This queue trades some features from VecDeque in the std library for the feature of direct indexing.
-Instead of returning `()`, [`push`] returns the index of the new element in the buffer. Methods like
-[`get_absolute`] can then be used for element access.
+Instead of returning `()`, [`push_back`] returns the index of the new element in the buffer. Methods
+like [`get_absolute`] can then be used for element access.
 
-The index returned by [`push`] or used by [`get_absolute`] may not be the actual index but a higher
-number that gets wrapped around internally so the indices are still valid after elements get moved
-around by expanding or shrinking the container.
+The index returned by [`push_back`] or used by [`get_absolute`] may not be the actual index but a
+higher number that gets wrapped around internally so the indices are still valid after elements get
+moved around by expanding or shrinking the container.
 
 The features not present in `RingBuffer` whereas present in `VecDeque` include:
 - Double ended functionality, including `push_front` and `pop_back`
@@ -36,14 +37,16 @@ The features not present in `RingBuffer` whereas present in `VecDeque` include:
 - `remove`
 - basically anything that messes with the position of elements in the buffer apart from `swap`
 
-[`push`]: Self::push
+[`push_back`]: Self::push_back
 [`get_absolute`]: Self::get_absolute
 */
 pub struct RingBuffer<T> {
     /// Pointer to the allocated array data.
-    memory: NonNull<T>,
+    data: NonNull<T>,
     /// Tells the compiler the struct owns `T` values.
     _marker: PhantomData<T>,
+    /// A bitfield signaling if a value got moved to this position by growing from a wrapped state.
+    moved_value: Box<[u64]>,
     /// The count of elements currently saved in the buffer.
     len: usize,
     /// The amount of element that fit into the buffer before reallocation.
@@ -54,6 +57,15 @@ pub struct RingBuffer<T> {
     tail: usize,
 }
 
+/// A direct index into the buffer.
+#[derive(Clone, Copy, Debug)]
+pub struct RingBufferIndex {
+    /// The index at the time of allocation.
+    base_index: usize,
+    /// An offset that is applied if the value got moved by growing the buffer.
+    wrap_offset: usize,
+}
+
 impl<T> RingBuffer<T> {
     /// Creates an empty `RingBuffer`.
     ///
@@ -85,23 +97,26 @@ impl<T> RingBuffer<T> {
             let layout =
                 Layout::from_size_align(capacity * size_of::<T>(), align_of::<T>()).unwrap();
             let ptr = unsafe { alloc(layout) };
-
             if ptr.is_null() {
                 handle_alloc_error(layout);
-            } else {
-                RingBuffer {
-                    memory: NonNull::new(ptr).unwrap().cast(),
-                    _marker: PhantomData::<T>,
-                    len: 0,
-                    capacity,
-                    head: 0,
-                    tail: 0,
-                }
+            }
+
+            RingBuffer {
+                data: NonNull::new(ptr).unwrap().cast(),
+                _marker: PhantomData::<T>,
+                moved_value: vec![0; (capacity + u64::BITS as usize - 1) / u64::BITS as usize]
+                    .into(),
+                len: 0,
+                capacity,
+                head: 0,
+                tail: 0,
             }
         } else {
             RingBuffer {
-                memory: NonNull::dangling(),
+                data: NonNull::dangling(),
                 _marker: PhantomData::<T>,
+                moved_value: vec![0; (capacity + u64::BITS as usize - 1) / u64::BITS as usize]
+                    .into(),
                 len: 0,
                 capacity: usize::MAX,
                 head: 0,
@@ -137,15 +152,16 @@ impl<T> RingBuffer<T> {
     ///
     /// assert_eq!(buffer.as_slices(), (&[1, 2][..], &[3, 4][..]));
     /// ```
+    #[inline]
     pub fn as_slices(&self) -> (&[T], &[T]) {
         unsafe {
             if self.wraps_around() {
                 (
                     slice::from_raw_parts(
-                        self.memory.as_ptr().add(self.tail),
+                        self.data.as_ptr().add(self.tail),
                         self.capacity - self.tail,
                     ),
-                    slice::from_raw_parts(self.memory.as_ptr(), self.head),
+                    slice::from_raw_parts(self.data.as_ptr(), self.head),
                 )
             } else {
                 let head_or_capacity = if self.head == 0 {
@@ -156,7 +172,7 @@ impl<T> RingBuffer<T> {
 
                 (
                     slice::from_raw_parts(
-                        self.memory.as_ptr().add(self.tail),
+                        self.data.as_ptr().add(self.tail),
                         head_or_capacity - self.tail,
                     ),
                     &[],
@@ -227,9 +243,12 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.get_absolute(back_index.unwrap()), buffer.back());
     /// ```
     #[inline]
-    pub fn back_absolute_index(&self) -> Option<usize> {
+    pub fn back_absolute_index(&self) -> Option<RingBufferIndex> {
         if self.len > 0 {
-            Some((self.head.wrapping_sub(1) & self.mask()) + self.capacity)
+            Some(RingBufferIndex {
+                base_index: self.head.wrapping_sub(1) & self.mask(),
+                wrap_offset: self.capacity * (self.wraps_around() & (self.head != 0)) as usize,
+            })
         } else {
             None
         }
@@ -268,6 +287,9 @@ impl<T> RingBuffer<T> {
         while self.len > 0 {
             self.pop_front();
         }
+        for field in self.moved_value.iter_mut() {
+            *field = 0;
+        }
         self.len = 0;
         self.head = 0;
         self.tail = 0;
@@ -361,9 +383,12 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.get_absolute(front_index.unwrap()), buffer.front());
     /// ```
     #[inline]
-    pub fn front_absolute_index(&self) -> Option<usize> {
+    pub fn front_absolute_index(&self) -> Option<RingBufferIndex> {
         if self.len > 0 {
-            Some(self.tail + self.capacity * self.wraps_around() as usize)
+            Some(RingBufferIndex {
+                base_index: self.tail,
+                wrap_offset: 0,
+            })
         } else {
             None
         }
@@ -372,11 +397,11 @@ impl<T> RingBuffer<T> {
     /// Retrieves an element in the `RingBuffer` by absolute index.
     ///
     /// The index provided is the index used by the internal container.
-    /// This is designed to be used in conjunction with [`push`].
+    /// This is designed to be used in conjunction with [`push_back`].
     /// This also gets wrapped around, supporting expanding and shrinking the container without
     /// breaking the access with absolute addresses.
     ///
-    /// [`push`]: Self::push
+    /// [`push_back`]: Self::push_back
     ///
     /// # Example
     ///
@@ -393,9 +418,9 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.get_absolute(second), Some(&42));
     /// ```
     #[inline]
-    pub fn get_absolute(&self, index: usize) -> Option<&T> {
+    pub fn get_absolute(&self, index: RingBufferIndex) -> Option<&T> {
         if self.is_index_in_range(index) {
-            unsafe { Some(&*self.memory.as_ptr().add(index & self.mask())) }
+            unsafe { Some(&*self.data.as_ptr().add(self.resolve_index(index))) }
         } else {
             None
         }
@@ -404,11 +429,11 @@ impl<T> RingBuffer<T> {
     /// Retrieves a mutable element in the `RingBuffer` by absolute index.
     ///
     /// The index provided is the index used by the internal container.
-    /// This is designed to be used in conjunction with [`push`].
+    /// This is designed to be used in conjunction with [`push_back`].
     /// This also gets wrapped around, supporting expanding and shrinking the container without
     /// breaking the access with absolute addresses.
     ///
-    /// [`push`]: Self::push
+    /// [`push_back`]: Self::push_back
     ///
     /// # Example
     ///
@@ -423,15 +448,16 @@ impl<T> RingBuffer<T> {
     ///
     /// assert_eq!(buffer.get_absolute(the_answer), Some(&2));
     ///
-    /// if let Some(x) = buffer.get_relative_mut(the_answer) {
+    /// if let Some(x) = buffer.get_absolute_mut(the_answer) {
     ///     *x = 42;
     /// }
     ///
     /// assert_eq!(buffer.get_absolute(the_answer), Some(&42));
     /// ```
-    pub fn get_absolute_mut(&mut self, index: usize) -> Option<&mut T> {
+    #[inline]
+    pub fn get_absolute_mut(&mut self, index: RingBufferIndex) -> Option<&mut T> {
         if self.is_index_in_range(index) {
-            unsafe { Some(&mut *self.memory.as_ptr().add(index & self.mask())) }
+            unsafe { Some(&mut *self.data.as_ptr().add(self.resolve_index(index))) }
         } else {
             None
         }
@@ -467,8 +493,8 @@ impl<T> RingBuffer<T> {
     /// }
     /// ```
     #[inline]
-    pub unsafe fn get_absolute_unchecked(&self, index: usize) -> &T {
-        &*self.memory.as_ptr().add(index & self.mask())
+    pub unsafe fn get_absolute_unchecked(&self, index: RingBufferIndex) -> &T {
+        &*self.data.as_ptr().add(self.resolve_index(index))
     }
 
     /// Retrieves a mutable element in the `RingBuffer` by absolute index without checking for
@@ -505,8 +531,8 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.get_absolute(the_answer), Some(&42));
     /// ```
     #[inline]
-    pub unsafe fn get_absolute_mut_unchecked(&mut self, index: usize) -> &mut T {
-        &mut *self.memory.as_ptr().add(index & self.mask())
+    pub unsafe fn get_absolute_mut_unchecked(&mut self, index: RingBufferIndex) -> &mut T {
+        &mut *self.data.as_ptr().add(self.resolve_index(index))
     }
 
     /// Retrieves an element in the `RingBuffer` by relative index.
@@ -610,13 +636,13 @@ impl<T> RingBuffer<T> {
     /// buffer.push_back(5);
     /// let b: &[_] = &[&3, &4, &5];
     /// let c: Vec<&usize> = buffer.iter().collect();
-    /// assert_eq!(buffer.get_absolute(0), Some(&3));
+    /// assert_eq!(buffer.front(), Some(&3));
     /// assert_eq!(&c[..], b);
     /// ```
     #[inline]
     pub fn iter(&self) -> Iter<T> {
         Iter {
-            buffer: unsafe { slice::from_raw_parts(self.memory.as_ptr(), self.capacity) },
+            buffer: unsafe { slice::from_raw_parts(self.data.as_ptr(), self.capacity) },
             len: self.len,
             index: self.tail,
         }
@@ -642,7 +668,7 @@ impl<T> RingBuffer<T> {
     #[inline]
     pub fn iter_mut(&mut self) -> IterMut<T> {
         IterMut {
-            buffer: unsafe { slice::from_raw_parts_mut(self.memory.as_ptr(), self.capacity) },
+            buffer: unsafe { slice::from_raw_parts_mut(self.data.as_ptr(), self.capacity) },
             len: self.len,
             index: self.tail,
         }
@@ -681,9 +707,10 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.pop_front(), Some(2));
     /// assert_eq!(buffer.pop_front(), None);
     /// ```
+    #[inline]
     pub fn pop_front(&mut self) -> Option<T> {
         if !self.is_empty() {
-            let return_value = unsafe { Some(self.memory.as_ptr().add(self.tail).read()) };
+            let return_value = unsafe { Some(self.data.as_ptr().add(self.tail).read()) };
 
             self.tail = (self.tail + 1) & self.mask();
             self.len -= 1;
@@ -705,20 +732,32 @@ impl<T> RingBuffer<T> {
     /// buffer.push_back(42);
     /// assert_eq!(42, *buffer.back().unwrap());
     /// ```
-    pub fn push_back(&mut self, new_element: T) -> usize {
+    #[inline]
+    pub fn push_back(&mut self, new_element: T) -> RingBufferIndex {
         if self.is_full() {
             self.grow(self.capacity << 1);
         }
 
         unsafe {
-            self.memory.as_ptr().add(self.head).write(new_element);
+            self.data.as_ptr().add(self.head).write(new_element);
         }
 
-        let return_index = self.head;
+        let wrap_offset;
+        if self.wraps_around() {
+            let check_bit = 1 << (self.head & (u64::BITS as usize - 1));
+            self.moved_value[self.head / u64::BITS as usize] |= check_bit;
+            wrap_offset = self.capacity;
+        } else {
+            wrap_offset = 0;
+        }
+        let base_index = self.head;
         self.head = (self.head + 1) & self.mask();
         self.len += 1;
 
-        return_index + (self.capacity * self.wraps_around() as usize)
+        RingBufferIndex {
+            base_index,
+            wrap_offset,
+        }
     }
 
     /// Reserves capacity for at least `additional` more elements to be inserted in the given
@@ -756,6 +795,9 @@ impl<T> RingBuffer<T> {
     /// Does nothing if the current capacity is smaller than the supplied
     /// minimum capacity.
     ///
+    /// Also only shrinks if no element position needs to change as to preserve the indices of the
+    /// allocated elements.
+    ///
     /// # Example
     ///
     /// ```
@@ -779,56 +821,42 @@ impl<T> RingBuffer<T> {
                 self.len.next_power_of_two(),
             ),
         );
-        if new_capacity < self.capacity {
+        if (new_capacity < self.capacity)
+            & !self.wraps_around()
+            & !self.is_empty()
+            & (self.head != 0)
+            & (self.head <= new_capacity)
+        {
             unsafe {
-                let layout =
-                    Layout::from_size_align(new_capacity * size_of::<T>(), align_of::<T>())
-                        .unwrap();
+                let layout = Layout::from_size_align_unchecked(
+                    new_capacity * size_of::<T>(),
+                    align_of::<T>(),
+                );
                 let new_ptr: *mut T = alloc(layout) as *mut T;
                 if new_ptr.is_null() {
                     handle_alloc_error(layout);
-                } else {
-                    let new_mask = new_capacity - 1;
-
-                    if self.wraps_around() {
-                        self.memory.as_ptr().copy_to(new_ptr, self.head);
-                        self.memory
-                            .as_ptr()
-                            .add(self.tail)
-                            .copy_to(new_ptr.add(self.tail & new_mask), self.capacity - self.tail);
-                    } else if !self.is_empty() & ((self.head > new_capacity) | (self.head == 0)) {
-                        if self.tail >= new_capacity {
-                            self.memory
-                                .as_ptr()
-                                .add(self.tail)
-                                .copy_to(new_ptr.add(self.tail & new_mask), self.len);
-                        } else {
-                            self.memory
-                                .as_ptr()
-                                .add(self.capacity)
-                                .copy_to(new_ptr, self.head);
-                            self.memory.as_ptr().add(self.tail).copy_to(
-                                new_ptr.add(self.tail & new_mask),
-                                self.capacity - self.tail,
-                            );
-                        }
-                    } else {
-                        self.memory
-                            .as_ptr()
-                            .add(self.tail)
-                            .copy_to(new_ptr.add(self.tail), self.len);
-                    }
-
-                    dealloc(
-                        self.memory.as_ptr() as *mut u8,
-                        Layout::from_size_align(self.capacity * size_of::<T>(), align_of::<T>())
-                            .unwrap(),
-                    );
-
-                    self.memory = NonNull::new_unchecked(new_ptr);
                 }
+
+                self.data
+                    .as_ptr()
+                    .add(self.tail)
+                    .copy_to(new_ptr.add(self.tail), self.len);
+
+                dealloc(
+                    self.data.as_ptr() as *mut u8,
+                    Layout::from_size_align(self.capacity * size_of::<T>(), align_of::<T>())
+                        .unwrap(),
+                );
+
+                self.data = NonNull::new_unchecked(new_ptr);
             }
 
+            self.moved_value = self
+                .moved_value
+                .iter()
+                .copied()
+                .take(new_capacity)
+                .collect();
             self.capacity = new_capacity;
             self.head &= self.mask();
             self.tail &= self.mask();
@@ -903,9 +931,9 @@ impl<T> RingBuffer<T> {
     ///
     /// `i` and `j` may be equal.
     ///
-    /// This is designed to be used in conjunction with [`push`].
+    /// This is designed to be used in conjunction with [`push_back`].
     ///
-    /// [`push`]: Self::push
+    /// [`push_back`]: Self::push_back
     ///
     /// # Panics
     ///
@@ -926,25 +954,25 @@ impl<T> RingBuffer<T> {
     /// assert_eq!(buffer.get_absolute(first), Some(&4));
     /// assert_eq!(buffer.get_absolute(second), Some(&6));
     ///
-    /// buffer.swap_relative(first, second);
+    /// buffer.swap_absolute(first, second);
     ///
     /// assert_eq!(buffer.get_absolute(first), Some(&6));
     /// assert_eq!(buffer.get_absolute(second), Some(&4));
     /// ```
-    pub fn swap_absolute(&mut self, i: usize, j: usize) {
+    pub fn swap_absolute(&mut self, i: RingBufferIndex, j: RingBufferIndex) {
         assert!(
             self.is_index_in_range(i) & self.is_index_in_range(j),
             "At least one index is out of bounds, i is {}, j is {}, buffer is from {} to {}",
-            i,
-            j,
+            self.resolve_index(i),
+            self.resolve_index(j),
             self.tail,
             self.head
         );
         unsafe {
-            self.memory
+            self.data
                 .as_ptr()
-                .add(i)
-                .swap(self.memory.as_ptr().add(j));
+                .add(self.resolve_index(i))
+                .swap(self.data.as_ptr().add(self.resolve_index(j)));
         }
     }
 
@@ -976,64 +1004,71 @@ impl<T> RingBuffer<T> {
     // private functions
     /// Tests if the given address is in the valid
     #[inline]
-    fn is_index_in_range(&self, index: usize) -> bool {
+    fn is_index_in_range(&self, index: RingBufferIndex) -> bool {
+        let base_index = self.resolve_index(index);
         if self.wraps_around() {
-            (index & self.mask()).wrapping_sub(self.head) >= (self.capacity - self.len)
+            base_index.wrapping_sub(self.head) >= (self.capacity - self.len)
         } else {
-            (index & self.mask()).wrapping_sub(self.tail) < self.len
+            base_index.wrapping_sub(self.tail) < self.len
         }
     }
 
     /// Expands the container to a larger one and copies the data. Unwraps the `RingBuffer` if it
     /// was wrapped before.
     fn grow(&mut self, new_capacity: usize) {
+        let mut moved_value = vec![0; (new_capacity + u64::BITS as usize - 1) / u64::BITS as usize];
+        moved_value[..(self.capacity + u64::BITS as usize - 1) / u64::BITS as usize]
+            .copy_from_slice(&self.moved_value[..]);
+
         unsafe {
             assert_ne!(size_of::<T>(), 0, "capacity overflow");
+            assert!(new_capacity.is_power_of_two());
 
             let layout =
-                Layout::from_size_align(new_capacity * size_of::<T>(), align_of::<T>()).unwrap();
+                Layout::from_size_align_unchecked(new_capacity * size_of::<T>(), align_of::<T>());
             let new_ptr: *mut T = alloc(layout) as *mut T;
             if new_ptr.is_null() {
                 handle_alloc_error(layout);
-            } else {
-                if self.wraps_around() {
-                    self.memory
-                        .as_ptr()
-                        .add(self.tail)
-                        .copy_to(new_ptr.add(self.tail), self.capacity - self.tail);
-                    self.memory
-                        .as_ptr()
-                        .copy_to(new_ptr.add(self.capacity), self.head);
-                    self.head += self.capacity;
-                } else if !self.is_empty() {
-                    self.memory
-                        .as_ptr()
-                        .add(self.tail)
-                        .copy_to(new_ptr.add(self.tail), self.len);
-                    // Edge case
-                    if self.head == 0 {
-                        self.head = self.capacity;
-                    }
-                }
-
-                dealloc(
-                    self.memory.as_ptr() as *mut u8,
-                    Layout::from_size_align(self.capacity * size_of::<T>(), align_of::<T>())
-                        .unwrap(),
-                );
-
-                self.memory = NonNull::new_unchecked(new_ptr);
             }
+
+            if self.wraps_around() {
+                self.data
+                    .as_ptr()
+                    .add(self.tail)
+                    .copy_to(new_ptr.add(self.tail), self.capacity - self.tail);
+                self.data
+                    .as_ptr()
+                    .copy_to(new_ptr.add(self.capacity), self.head);
+                self.head += self.capacity;
+
+                for i in self.capacity..self.head {
+                    let check_bit = 1 << (i & (u64::BITS as usize - 1));
+                    moved_value[i / u64::BITS as usize] |= check_bit;
+                }
+            } else if !self.is_empty() {
+                self.data
+                    .as_ptr()
+                    .add(self.tail)
+                    .copy_to(new_ptr.add(self.tail), self.len);
+            }
+
+            dealloc(
+                self.data.as_ptr() as *mut u8,
+                Layout::from_size_align(self.capacity * size_of::<T>(), align_of::<T>()).unwrap(),
+            );
+
+            self.data = NonNull::new_unchecked(new_ptr);
         }
 
         self.capacity = new_capacity;
+        self.moved_value = moved_value.into();
     }
 
     /// A more convenient way to access a relative addressed element. Public access method only
     /// need to add their sugar.
     #[inline]
     unsafe fn get_relative_pointer(&self, index: usize) -> *mut T {
-        self.memory
+        self.data
             .as_ptr()
             .add(self.tail.wrapping_add(index) & self.mask())
     }
@@ -1044,10 +1079,23 @@ impl<T> RingBuffer<T> {
         self.capacity - 1
     }
 
+    /// Converts a [`RingBufferIndex`] into its effective index.
+    #[inline]
+    fn resolve_index(&self, index: RingBufferIndex) -> usize {
+        if index.wrap_offset < self.capacity {
+            let check_index = index.base_index + index.wrap_offset;
+            let check_bit = 1 << (check_index as u64 & (u64::BITS as u64 - 1));
+            let apply_offset = (self.moved_value[check_index / u64::BITS as usize] & check_bit) > 0;
+            index.base_index + index.wrap_offset * apply_offset as usize
+        } else {
+            index.base_index
+        }
+    }
+
     /// If the container wraps around.
     #[inline]
     fn wraps_around(&self) -> bool {
-        !self.is_empty() & (self.head != 0) & (self.head <= self.tail)
+        !self.is_empty() & (self.head <= self.tail)
     }
 }
 
@@ -1085,7 +1133,7 @@ impl<T> Drop for RingBuffer<T> {
         if size_of::<T>() > 0 {
             unsafe {
                 dealloc(
-                    self.memory.as_ptr() as *mut u8,
+                    self.data.as_ptr() as *mut u8,
                     Layout::from_size_align_unchecked(
                         self.capacity * size_of::<T>(),
                         align_of::<T>(),
@@ -1161,18 +1209,18 @@ impl<T> FromIterator<T> for RingBuffer<T> {
     }
 }
 
-impl<T> Index<usize> for RingBuffer<T> {
+impl<T> Index<RingBufferIndex> for RingBuffer<T> {
     type Output = T;
 
     #[inline]
-    fn index(&self, index: usize) -> &T {
+    fn index(&self, index: RingBufferIndex) -> &T {
         self.get_absolute(index).expect("Out of bounds access")
     }
 }
 
-impl<T> IndexMut<usize> for RingBuffer<T> {
+impl<T> IndexMut<RingBufferIndex> for RingBuffer<T> {
     #[inline]
-    fn index_mut(&mut self, index: usize) -> &mut T {
+    fn index_mut(&mut self, index: RingBufferIndex) -> &mut T {
         self.get_absolute_mut(index).expect("Out of bounds access")
     }
 }
@@ -1246,6 +1294,15 @@ where
 unsafe impl<T> Send for RingBuffer<T> where T: Send {}
 unsafe impl<T> Sync for RingBuffer<T> where T: Sync {}
 
+impl PartialEq for RingBufferIndex {
+    /// This is partially flawed as indices can be aliased in a way that two indices point to the
+    /// same element. The [`RingBuffer`] struct is needed to resolve the aliasing, but it's not
+    /// available here.
+    fn eq(&self, other: &Self) -> bool {
+        (self.base_index + self.wrap_offset) == (other.base_index + other.wrap_offset)
+    }
+}
+
 /// An iterator over the elements of a `RingBuffer`.
 ///
 /// This `struct` is created by the [`iter`] method on [`RingBuffer`]. See its
@@ -1432,7 +1489,7 @@ impl<T> FusedIterator for IntoIter<T> {}
 
 #[cfg(test)]
 mod tests {
-    use crate::RingBuffer;
+    use crate::{RingBuffer, RingBufferIndex};
 
     #[test]
     fn access_and_reallocations() {
@@ -1440,11 +1497,24 @@ mod tests {
         let mut indices = Vec::with_capacity(8);
         assert_eq!(buffer.capacity, 2, "Initialised with too much capacity");
 
-        // Phase 1: Filling the buffer and trigger extension
+        // Phase 1: Filling the buffer and trigger extensions
         for i in 1..=8 {
             indices.push(buffer.push_back(i));
         }
-        assert_eq!(indices.as_slice(), &[0, 1, 2, 3, 4, 5, 6, 7][..]);
+        assert_eq!(
+            indices.first().unwrap(),
+            &RingBufferIndex {
+                base_index: 0,
+                wrap_offset: 0
+            }
+        );
+        assert_eq!(
+            indices.last().unwrap(),
+            &RingBufferIndex {
+                base_index: 7,
+                wrap_offset: 0
+            }
+        );
         assert_eq!(buffer.capacity, 8);
         assert_eq!(buffer.as_slices(), (&[1, 2, 3, 4, 5, 6, 7, 8][..], &[][..]));
         for i in 0..8 {
@@ -1455,67 +1525,76 @@ mod tests {
         // Phase 2: Wrap around and trigger extension
         indices.clear();
         for i in 1..=4 {
-            indices.push(i + 3);
+            indices.push(RingBufferIndex {
+                base_index: i + 3,
+                wrap_offset: 0,
+            });
             assert_eq!(buffer.pop_front(), Some(i));
         }
+
         for i in 9..=12 {
             indices.push(buffer.push_back(i));
         }
-        assert_eq!(indices.as_slice(), &[4, 5, 6, 7, 8, 9, 10, 11][..]);
+        assert_eq!(
+            indices[0],
+            RingBufferIndex {
+                base_index: 4,
+                wrap_offset: 0
+            }
+        );
+        assert_eq!(
+            indices[3],
+            RingBufferIndex {
+                base_index: 7,
+                wrap_offset: 0
+            }
+        );
+        assert_eq!(
+            indices[4],
+            RingBufferIndex {
+                base_index: 0,
+                wrap_offset: 8
+            }
+        );
+        assert_eq!(
+            indices[7],
+            RingBufferIndex {
+                base_index: 3,
+                wrap_offset: 8
+            }
+        );
         assert_eq!(buffer.capacity, 8);
         assert_eq!(
             buffer.as_slices(),
             (&[5, 6, 7, 8][..], &[9, 10, 11, 12][..])
         );
+
         for i in 13..=16 {
             indices.push(buffer.push_back(i));
         }
         assert_eq!(
-            indices.as_slice(),
-            &[4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15][..]
+            indices[8],
+            RingBufferIndex {
+                base_index: 12,
+                wrap_offset: 0
+            }
+        );
+        assert_eq!(
+            indices[11],
+            RingBufferIndex {
+                base_index: 15,
+                wrap_offset: 0
+            }
         );
         assert_eq!(buffer.capacity, 16);
         assert_eq!(
             buffer.as_slices(),
             (&[5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16][..], &[][..])
         );
+
         for i in 0..12 {
             assert_eq!(buffer.get_relative(i), Some(&(i + 5)));
             assert_eq!(buffer.get_absolute(indices[i]), Some(&(i + 5)));
         }
-
-        // Phase 3: Shrink, wrap around and shrink again
-        indices.clear();
-        for i in 5..=8 {
-            indices.push(i + 3);
-            assert_eq!(buffer.pop_front(), Some(i));
-        }
-        for i in 12..=15 {
-            indices.push(i);
-        }
-        buffer.shrink_to_fit();
-        assert_eq!(indices.as_slice(), &[8, 9, 10, 11, 12, 13, 14, 15][..]);
-        assert_eq!(buffer.capacity, 8);
-        assert_eq!(
-            buffer.as_slices(),
-            (&[9, 10, 11, 12, 13, 14, 15, 16][..], &[][..])
-        );
-        indices.clear();
-        for i in 9..=14 {
-            assert_eq!(buffer.pop_front(), Some(i));
-        }
-        indices.push(14);
-        indices.push(15);
-        for i in 17..=18 {
-            indices.push(buffer.push_back(i));
-        }
-        buffer.shrink_to_fit();
-        assert_eq!(indices.as_slice(), &[14, 15, 8, 9][..]);
-        assert_eq!(buffer.capacity, 4);
-        assert_eq!(buffer.as_slices(), (&[15, 16][..], &[17, 18][..]));
-        for i in 0..4 {
-            assert_eq!(buffer.get_relative(i), Some(&(i + 15)));
-            assert_eq!(buffer.get_absolute(indices[i]), Some(&(i + 15)));
-        }
     }
 }
